# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Upgrade Views for ASGI
"""

import asyncio
import json
import os
from urllib.parse import parse_qs

from tailbone.views.asgi import WebsocketView
from tailbone.progress import get_basic_session


class UpgradeWS(WebsocketView):

    async def execution_progress(self, scope, receive, send):
        rattail_config = self.registry['rattail_config']

        # is user allowed to see this?
        if not await self.authorize(scope, receive, send, 'upgrades.execute'):
            return

        # this tracks when client disconnects
        state = {'disconnected': False}

        async def wait_for_disconnect():
            message = await receive()
            if message['type'] == 'websocket.disconnect':
                state['disconnected'] = True

        # watch for client disconnect, while we do other things
        asyncio.create_task(wait_for_disconnect())

        query = scope['query_string'].decode('utf_8')
        query = parse_qs(query)
        uuid = query['uuid'][0]
        progress_session_id = 'upgrades.{}.execution_progress'.format(uuid)
        progress_session = get_basic_session(rattail_config,
                                             id=progress_session_id)

        # do the rest forever, until client disconnects
        while not state['disconnected']:

            # load latest progress data
            progress_session.load()

            # when upgrade progress is complete...
            if progress_session.get('complete'):

                # maybe set success flash msg
                msg = progress_session.get('success_msg')
                if msg:
                    user_session = await self.get_user_session(scope)
                    user_session.flash(msg)
                    user_session.persist()

                # tell client progress is complete
                await send({'type': 'websocket.send',
                            'subtype': 'upgrades.execute_progress',
                            'text': json.dumps({'complete': True})})

                # this websocket is done
                break

            # we will send this data down to client
            data = dict(progress_session)

            # maybe add more lines from command output
            path = rattail_config.upgrade_filepath(uuid, filename='stdout.log')
            offset = progress_session.get('stdout.offset', 0)
            if os.path.exists(path):
                size = os.path.getsize(path) - offset
                if size > 0:
                    with open(path, 'rb') as f:
                        f.seek(offset)
                        chunk = f.read(size)
                        data['stdout'] = chunk.decode('utf8').replace('\n', '<br />')
                    progress_session['stdout.offset'] = offset + size
                    progress_session.save()

            # send data to client
            await send({'type': 'websocket.send',
                        'subtype': 'upgrades.execute_progress',
                        'text': json.dumps(data)})

            # pause for 1 second
            await asyncio.sleep(1)

    @classmethod
    def defaults(cls, config):
        cls._defaults(config)

    @classmethod
    def _defaults(cls, config):

        # execution progress
        config.add_tailbone_websocket('upgrades.execution_progress',
                                      cls, attr='execution_progress')


def defaults(config, **kwargs):
    base = globals()

    UpgradeWS = kwargs.get('UpgradeWS', base['UpgradeWS'])
    UpgradeWS.defaults(config)


def includeme(config):
    defaults(config)
